"""
Time Serie manipulation
"""
from typing import Dict, List, Optional

import numpy as np
import pandas as pd

from . import api
from . import search
from .common import to_datetime


def time_serie(ts_code: int, start: str, end: str, strict: bool = False) -> pd.Series:
    """
    Request a time serie data.

    :param ts_code: time serie code.
    :param start: start date (DD/MM/YYYY).
    :param end: end date (DD/MM/YYYY).
    :param strict: boolean to enforce a strict date range.

    :return: Time serie values as pandas Series indexed by date.
    :rtype: pandas.Series_

    Usage::

        >>> CDI = 12
        >>> ts = sgs.time_serie(CDI, start='02/01/2018', end='31/12/2018')
        >>> ts.head()
        2018-01-02    0.026444
        2018-01-03    0.026444
        2018-01-04    0.026444
        2018-01-05    0.026444
        2018-01-08    0.026444
    """
    
    if strict:
        ts_data = api.get_data_with_strict_range(ts_code, start, end)
    else:
        ts_data = api.get_data(ts_code, start, end)
        
    values = []
    index = []
    for i in ts_data:
        values.append(i["valor"])
        index.append(to_datetime(i["data"], "pt"))

    # Transform empty strings in null values
    values = [np.nan if value == "" else value for value in values]
    return pd.Series(values, index, name=ts_code, dtype=np.float)



from typing import Optional, Dict, List, Union

import pandas as pd

from .search import search_ts


def metadata(ts_code: Union[int, pd.DataFrame], language: str = "en") -> Optional[List]:
    """Request metadata about a time serie or all time series in a pandas dataframe.

    :param ts_code: time serie code or pandas dataframe with time series as columns.
    :param language: language of the returned metadata.

    :return: List of dicts containing time series metadata.
    :rtype: list_

    Usage::

        >>> CDI = 12
        >>> INCC = 192  #  National Index of Building Costs
        >>> df = sgs.dataframe([CDI, INCC], start='02/01/2018', end='31/12/2018')
        >>> sgs.metadata(df)
        [{'code': 12, 'name': 'Interest rate - CDI', 'unit': '% p.d.', 'frequency': 'D',
        'first_value': Timestamp('1986-03-06 00:00:00'), 'last_value': Timestamp('2019-06-27 00:00:00'),
        'source': 'Cetip'}, {'code': 192, 'name': 'National Index of Building Costs (INCC)',
        'unit': 'Monthly % var.', 'frequency': 'M', 'first_value': Timestamp('1944-02-29 00:00:00'),
        'last_value': Timestamp('2019-05-01 00:00:00'), 'source': 'FGV'}]
    """
    info = []
    if isinstance(ts_code, pd.core.frame.DataFrame):
        for col in ts_code.columns:
            col_info = search_ts(col, language)
            if col_info is not None:
                info.append(col_info[0])
            else:
                info.append(None)
    else:
        col_info = search_ts(ts_code, language)
        info.append(col_info)
    return info



"""
Shared functions.
"""
from datetime import datetime
import locale
import re
from typing import Union
import os


LRU_CACHE_SIZE = 32
MAX_ATTEMPT_NUMBER = 5


def to_datetime(date_string: str, language: str) -> Union[datetime, str]:
    """ Converts a date string to a datetime object """
    locales = {"pt": "pt_BR.utf-8", "en": "en_US.utf-8"}

    """ correct problem with locale in Windows platform """
    if os.name == 'nt':
        locales = {"pt": "Portuguese_Brazil.1252", "en": "Portuguese_Brazil.1252"}
    
    locale.setlocale(locale.LC_TIME, locales[language])

    dd_mm_aaaa = "%d/%m/%Y"
    mmm_aaaa = "%b/%Y"

    formats = [dd_mm_aaaa, mmm_aaaa]

    for fmt in formats:
        try:
            date = datetime.strptime(date_string, fmt)
            break
        except ValueError:
            continue
    else:
        yyyy = "[0-9]{4}"
        if re.match(yyyy, date_string):
            year = int(date_string)
            month = 12
            day = 31
            date = datetime(year, month, day)
        else:
            return date_string  # returns original value if cant parse

    return date



import functools
from typing import Union, List, Dict

import pandas as pd
import requests
from retrying import retry

from .common import LRU_CACHE_SIZE, MAX_ATTEMPT_NUMBER, to_datetime


@retry(stop_max_attempt_number=MAX_ATTEMPT_NUMBER)
@functools.lru_cache(maxsize=LRU_CACHE_SIZE)
def get_data(ts_code: int, begin: str, end: str) -> List:
    """
    Requests time series data from the SGS API in json format.
    """

    url = (
        "http://api.bcb.gov.br/dados/serie/bcdata.sgs.{}"
        "/dados?formato=json&dataInicial={}&dataFinal={}"
    )
    request_url = url.format(ts_code, begin, end)
    response = requests.get(request_url)
    return response.json()
    
def get_data_with_strict_range(ts_code: int, begin: str, end: str) -> List:

    """
    Request time series data from the SGS API considering a strict range of dates.
        
    SGS API default behaviour returns the last stored value when selected date range have no data.

    It is possible to catch this behaviour when the first record date precedes the start date.
    
    This function enforces an empty data set when the first record date precedes the start date, avoiding records out of selected range.
    
    :param ts_code: time serie code.
    :param begin: start date (DD/MM/YYYY).
    :param end: end date (DD/MM/YYYY).
  
    :return: Data in json format or an empty list
    :rtype: list

    """
    data = get_data(ts_code, begin, end)
    
    first_record_date = to_datetime(data[0]["data"], "pt")
    period_start_date = to_datetime(begin, 'pt')
    
    try:
        is_out_of_range =  first_record_date < period_start_date  #type: ignore
        if is_out_of_range:
            raise ValueError
    except TypeError:
        print("ERROR: Serie " + str(ts_code) + " - Please, use 'DD/MM/YYYY' format for date strings.")
        data = []
    except ValueError:
        print("WARNING: Serie " + str(ts_code) + " - There is no data for the requested period, but there's previous data.")
        data = []
    
    return data
